#!/usr/bin/env python

"""
A polar decoder class. Currently only Successive Cancellation Decoder (SCD) is supported.
"""

import numpy as np
from polarcodes.Math import Math
from timeit import default_timer as timer

class Decode(Math):
    def __init__(self, myPC, decoder_name = 'scd'):
        """
        :param myPC: a polar code object created using the :class:`PolarCode` class
        :param decoder_name: name of decoder to use (default is 'scd')
        :type myPC: :class:`PolarCode`
        :type decoder_name: string
        """

        self.myPC = myPC

        # select decoding algorithm
        if decoder_name == 'scd':
            self.polar_decode(self.myPC.likelihoods)

    def upper_llr(self, l1, l2):
        """
        Update top branch LLR in the log-domain.
        This function supports shortening by checking for infinite LLR cases.

        :param l1: input LLR corresponding to the top branch
        :param l2: input LLR corresponding to the bottom branch
        :type l1: float
        :type l2: float
        :return: the top branch LLR at the next stage of the decoding tree
        :rtype: float
        """

        # check for infinite LLR, used in shortening
        if l1 == np.inf and l2 != np.inf:
                return l2
        elif l1 != np.inf and l2 == np.inf:
            return l1
        elif l1 == np.inf and l2 == np.inf:
            return np.inf
        else:
            # principal decoding equation
            return self.logdomain_sum(l1 + l2, 0) - self.logdomain_sum(l1, l2)

    def lower_llr(self, l1, l2, b):
        """
        Update bottom branch LLR in the log-domain.
        This function supports shortening by checking for infinite LLR cases.

        :param l1: input LLR corresponding to the top branch
        :param l2: input LLR corresponding to the bottom branch
        :param b: the decoded bit of the top branch
        :type l1: float
        :type l2: float
        :type b: int
        :return: the bottom branch LLR at the next stage of the decoding tree
        :rtype: float or np.nan
        """

        if b == 0:
            # check for infinite LLRs, used in shortening
            if l1 == np.inf or l2 == np.inf:
                return np.inf
            else:
                # principal decoding equation
                return l1 + l2
        elif b == 1:
            # Principal decoding equation
            return l1 - l2
        return np.nan

    def update_LLR(self, x_ind):
        """
        Update all possible likelihoods at stage (n-j).
        This is a non-recursive implementation of :func:`update_LLR_recursive`.

        :param x_ind: the root index of a tree in the decoder graph
        :type x_ind: ind
        """

        for j in range(self.myPC.n - 1, -1, -1):
            s = 2 ** (self.myPC.n - j)  # partition length
            half_s = int(s / 2)  # half partition length

            for i in range(x_ind, self.myPC.N, s):  # do not need to update indices less than x_ind
                l = i % s
                if l < half_s:  # upper branch
                    l1 = self.L[i, j + 1]
                    l2 = self.L[i + half_s, j + 1]
                    self.L[i, j] = self.upper_llr(self.L[i, j + 1], l2)
                else:  # lower branch
                    l1 = self.L[i, j + 1]
                    l2 = self.L[i - half_s, j + 1]
                    b = self.B[i - half_s, j]
                    self.L[i, j] = self.lower_llr(l1, l2, b)

    def update_LLR_recursive(self, i, j):
        """
        Update all possible likelihoods at stage (n-j).
        This is a recursive implementation of :func:`update_LLR`.
        """

        s = 2 ** (self.myPC.n - j)  # partition size
        half_s = int(s / 2)
        l = i % s

        if l < half_s:  # upper branch
            if np.isnan(self.L[i, j + 1]):
                self.update_LLR_recursive(i, j + 1)
            if np.isnan(self.L[i + half_s, j + 1]):
                self.update_LLR_recursive(i + half_s, j + 1)

            # evaluate the f function
            l1 = self.L[i, j + 1]
            l2 = self.L[i + half_s, j + 1]
            self.L[i, j] = self.upper_llr(self.L[i, j + 1], l2)
        else:  # lower branch
            # evaluate the g function
            l1 = self.L[i, j + 1]
            l2 = self.L[i - half_s, j + 1]
            b = self.B[i - half_s, j]
            self.L[i, j] = self.lower_llr(l1, l2, b)

    def update_bits(self, x_ind):
        """
        Update all possible bits at stage (n-j).
        This is a non-recursive implementation.

        :param x_ind: the root index of a tree in the decoder graph
        :type x_ind: ind
        """

        next_i = [x_ind]  # store branch indices to update in next stage
        for j in range(self.myPC.n):
            s = 2 ** (self.myPC.n - j)  # partition length
            half_s = int(s / 2)  # half partition length

            next_i_temp = []
            for i in next_i:
                l = i % s
                if l >= half_s:  # lower branch
                    # propagate upper branch bit
                    self.B[i - half_s, j + 1] = int(self.B[i, j]) ^ int(self.B[i - half_s, j])
                    self.B[i, j + 1] = self.B[i, j]

                    # append branches to search in next stage
                    next_i_temp.append(i)
                    next_i_temp.append(i - half_s)
                next_i = next_i_temp  # update search array

    def polar_decode(self, y):
        """
        Successive Cancellation Decoder. The decoded message is set to ``message_received`` in ``myPC``.
        The decoder will use the frozen set as defined by ``frozen`` in ``myPC``.
        Depends on :func:`update_LLR_recursive` and :func:`update_bits`.

        -------------
        **References:**

        *  Vangala, H., Viterbo, & Yi Hong. (2014). Permuted successive cancellation decoder for polar codes. 2014 International Symposium on Information Theory and Its Applications, 438–442. IEICE.

        :param y: a vector of likelihoods at the channel output
        :type y: ndarray<float>
        """

        # decoding initial params
        self.L = np.full((self.myPC.N, self.myPC.n + 1), np.nan, dtype=np.float64)
        self.B = np.full((self.myPC.N, self.myPC.n + 1), np.nan)
        self.L[:, self.myPC.n] = y

        # decode rung by rung
        for i in range(0, self.myPC.N):
            # evaluate tree of LLRs for bit i
            l = self.bit_reversed(i, self.myPC.n)  # use bit reversal of i for the "natural order"
            self.update_LLR_recursive(l, 0)

            # make hard decision at output (first column of B)
            if l in self.myPC.frozen:
                self.B[l, 0] = 0
            else:
                if self.L[l, 0] >= 0:
                    self.B[l, 0] = 0
                else:
                    self.B[l, 0] = 1

            # propagate the hard decision just made
            self.update_bits(l)

        x_noisy = self.B[:, 0].astype(int)
        self.myPC.message_received = x_noisy[self.myPC.frozen_lookup == 1]



#!/usr/bin/env python

"""
Math provides miscellaneous math operations to the other classes, which are important for polar codes
algorithm implementations.
"""

import numpy as np

class Math:
    def bit_reversed(self, x, n):
        """
        Bit-reversal operation.

        :param x: a vector of indices
        :param n: number of bits per index in ``x``
        :type x: ndarray<int> or int
        :type n: int
        :return: bit-reversed version of x
        :rtype: ndarray<int> or int
        """

        result = 0
        for i in range(n):  # for each bit number
            if (x & (1 << i)):  # if it matches that bit
                result |= (1 << (n - 1 - i))  # set the "opposite" bit in result
        return result

    def logdomain_diff(self, x, y):
        """
        Find the difference between x and y in log-domain. It uses log1p to improve numerical stability.

        :param x: any number in the log-domain
        :param y: any number in the log-domain
        :type x: float
        :type y: float
        :return: the result of x - y
        :rtype: float
        """

        if x > y:
            z = x + np.log1p(-np.exp(y - x))
        else:
            z = y + np.log1p(-np.exp(x - y))
        return z

    def logdomain_sum(self, x, y):
        """
        Find the addition of x and y in log-domain. It uses log1p to improve numerical stability.

        :param x: any number in the log-domain
        :param y: any number in the log-domain
        :type x: float
        :type y: float
        :return: the result of x + y
        :rtype: float
        """

        if x > y:
            z = x + np.log1p(np.exp(y - x))
        else:
            z = y + np.log1p(np.exp(x - y))
        return z

    def bit_perm(self, x, p, n):
        """
        Find the permutation of an index.

        :param x: a vector of indices
        :param p: permutation vector, ex: bit-reversal is (0,1,...,n-1)
        :param n: number of bits per index in ``x``
        :type x: ndarray<int> or int
        :type p: ndarray<int>
        :type n: int
        :return: permuted indices
        :rtype: ndarray<int> or int
        """

        result = 0
        for i in range(n):
            b = (x >> p[i]) & 1
            result ^= (-b ^ result) & (1 << (n - i - 1))
        return result

    # find hamming weight of an index x
    def hamming_wt(self, x, n):
        """
        Find the bit-wise hamming weight of an index.

        :param x: an index
        :param n: number of bits in ``x``
        :type x: int
        :type n: int
        :return: bit-wise hamming weight of ``x``
        :rtype: int
        """

        m = 1
        wt = 0
        for i in range(n):
            b = (x >> i) & m
            if b:
                wt = wt + 1
        return wt

    # sort by hamming_wt()
    def sort_by_wt(self, x, n):
        """
        Sort a vector by index hamming weights using hamming_wt().

        :param x: a vector of indices
        :param n: number of bits per index in ``x``
        :type x: ndarray<int>
        :type n: int
        :return: sorted vector
        :rtype: ndarray<int>
        """

        wts = np.zeros(len(x), dtype=int)
        for i in range(len(x)):
            wts[i] = self.hamming_wt(x[i], n)
        mask = np.argsort(wts)
        return x[mask]

    def inverse_set(self, F, N):
        """
        Find {0,1,...,N-1}\F. This is useful for finding the information set given a frozen set as ``F``.

        :param F: a vector of indices
        :param N: block length
        :type F: ndarray<int>
        :type N: int
        :return: inverted set as a vector
        :rtype: ndarray<int>
        """

        n = int(np.log2(N))
        not_F = []
        for i in range(N):
            if i not in F:
                not_F.append(i)
        return np.array(not_F)

    def subtract_set(self, X, Y):
        """
        Subtraction of two sets.

        :param X: a vector of indices
        :param Y: a vector of indices
        :type X: ndarray<int>
        :type Y: ndarray<int>
        :return: result of subtracted set as a vector
        :rtype: ndarray<int>
        """

        X_new = []
        for x in X:
            if x not in Y:
                X_new.append(x)
        return np.array(X_new)

    def arikan_gen(self, n):
        """
        The n-th kronecker product of [[1, 1], [0, 1]], commonly referred to as Arikan's kernel.

        :param n: log2(N), where N is the block length
        :type n: int
        :return: polar code generator matrix
        :rtype: ndarray<int>
        """

        F = np.array([[1, 1], [0, 1]])
        F_n = F
        for i in range(n - 1):
            F_n = np.kron(F, F_n)
        return F_n

    # Gaussian Approximation helper functions:

    def phi_residual(self, x, val):
        return self.phi(x) - val

    def phi(self, x):
        if x < 10:
            y = -0.4527 * (x ** 0.86) + 0.0218
            y = np.exp(y)
        else:
            y = np.sqrt(3.14159 / x) * (1 - 10 / (7 * x)) * np.exp(-x / 4)
        return y

    def phi_inv(self, y):
        return self.bisection(y, 0, 10000)

    def bisection(self, val, a, b):
        c = a
        while (b - a) >= 0.01:
            # check if middle point is root
            c = (a + b) / 2
            if (self.phi_residual(c, val) == 0.0):
                break

            # choose which side to repeat the steps
            if (self.phi_residual(c, val) * self.phi_residual(a, val) < 0):
                b = c
            else:
                a = c
        return c

    def logQ_Borjesson(self, x):
        a = 0.339
        b = 5.510
        half_log2pi = 0.5 * np.log(2 * np.pi)
        if x < 0:
            x = -x
            y = -np.log((1 - a) * x + a * np.sqrt(b + x * x)) - (x * x / 2) - half_log2pi
            y = np.log(1 - np.exp(y))
        else:
            y = -np.log((1 - a) * x + a * np.sqrt(b + x * x)) - (x * x / 2) - half_log2pi
        return y



#!/usr/bin/env python

"""
Construct performs the mothercode construction.
It uses the algorithm specified by ``construction_type`` in ``myPC``.
Mothercode constructions supported: Bhattacharyya Bounds, Gaussian Approximation.
"""

import numpy as np
from polarcodes.Math import Math

class Construct(Math):
    def __init__(self, myPC, design_SNR, manual=False):
        """
        :param myPC: a polar code object created using the :class:`PolarCode` class
        :param design_SNR: the design SNR in decibels
        :param manual: suppress the constructor init
        :type myPC: :class:`PolarCode`
        :type design_SNR: float
        :type manual: bool
        """

        if manual:
            return
        else:
            self.update_mpcc(myPC, design_SNR)

    def update_mpcc(self, myPC, design_SNR):
        # select the mothercode construction method
        design_SNR_normalised = myPC.get_normalised_SNR(design_SNR)
        if myPC.construction_type == 'bb':
            z0 = -design_SNR_normalised
            myPC.reliabilities, myPC.frozen = self.general_pcc(myPC, z0)
        elif myPC.construction_type == 'ga':
            z0 = np.array([4 * design_SNR_normalised] * myPC.N)
            myPC.reliabilities, myPC.frozen = self.general_ga(myPC, z0)
        myPC.frozen_lookup = myPC.get_lut(myPC.frozen)

    def general_pcc(self, myPC, z0):
        """
        Polar code construction using Bhattacharyya Bounds. Each bit-channel can have different parameters.
        Supports shortening by adding extra cases for infinite likelihoods.

        -------------
        **References:**

        * Vangala, H., Viterbo, E., & Hong, Y. (2015). A Comparative Study of Polar Code Constructions for the AWGN Channel. arXiv.org. Retrieved from http://search.proquest.com/docview/2081709282/

        :param z0: a vector of the initial Bhattacharyya parameters in the log-domain, -E_b/N_o.
                    Note that this SNR should be normalised using :func:`get_normalised_SNR` in :class:`PolarCode`
        :type z0: ndarray<float> or float
        :return: channel reliabilities in log-domain (least reliable first), and the frozen indices
        :rtype: ndarray<int>, ndarray<int>
        """

        n = myPC.n
        z = np.zeros((myPC.N, n + 1))
        z[:, 0] = z0  # initial channel states

        for j in range(1, n + 1):
            u = 2 ** j  # number of branches at depth j
            for t in range(0, myPC.N, u):  # loop over top branches at this stage
                for s in range(int(u / 2)):
                    k = t + s
                    z_top = z[k, j - 1]
                    z_bottom = z[k + int(u / 2), j - 1]

                    # shortening infinity cases
                    if z_top == -np.inf and z_bottom != -np.inf:
                        z[k, j] = z_bottom
                        z[k + int(u / 2), j] = -np.inf
                    elif z_top != -np.inf and z_bottom == -np.inf:
                        z[k, j] = z_top
                        z[k + int(u / 2), j] = -np.inf
                    elif z_top == -np.inf and z_bottom == -np.inf:
                        z[k, j] = -np.inf
                        z[k + int(u / 2), j] = -np.inf
                    # principal equations
                    else:
                        z[k, j] = self.logdomain_diff(self.logdomain_sum(z_top, z_bottom), z_top + z_bottom)
                        z[k + int(u / 2), j] = z_top + z_bottom

        reliabilities = np.argsort(-z[:, n], kind='mergesort')   # ordered by least reliable to most reliable
        frozen = np.argsort(z[:, n], kind='mergesort')[myPC.K:]     # select N-K least reliable channels
        return reliabilities, frozen

    def perfect_pcc(self, myPC, p):
        """
        Boolean expression approach to puncturing pattern construction.

        -------------
        **References:**

        * Song-Nam, H., & Hui, D. (2018). On the Analysis of Puncturing for Finite-Length Polar Codes: Boolean Function Approach. arXiv.org. Retrieved from http://search.proquest.com/docview/2071252269/

        :param p: lookup table for coded puncturing bits. "0" => punctured, "1" => information.
                    For shortening, take the complement of p.
        :type p: ndarray<int>
        :return: a lookup table for which uncoded bits should be punctured.
            For shortening, take the complement of this output.
        :rtype: ndarray<int>
        """

        n = int(np.log2(myPC.N))
        z = np.zeros((myPC.N, n + 1), dtype=int)
        z[:, 0] = p

        for j in range(1, n + 1):
            u = 2 ** j  # number of branches at depth j
            # loop over top branches at this stage
            for t in range(0, myPC.N, u):
                for s in range(int(u / 2)):
                    k = t + s
                    z_top = z[k, j - 1]
                    z_bottom = z[k + int(u / 2), j - 1]
                    z[k, j] = z_top & z_bottom
                    z[k + int(u / 2), j] = z_top | z_bottom
        return z[:, n]

    def general_ga(self, myPC, z0):
        """
        Polar code construction using density evolution with the Gaussian Approximation. Each channel can have different parameters.

        -------------
        **References:**

        * Trifonov, P. (2012). Efficient Design and Decoding of Polar Codes. IEEE Transactions on Communications, 60(11), 3221–3227. https://doi.org/10.1109/TCOMM.2012.081512.110872

        * Vangala, H., Viterbo, E., & Hong, Y. (2015). A Comparative Study of Polar Code Constructions for the AWGN Channel. arXiv.org. Retrieved from http://search.proquest.com/docview/2081709282/

        :param z0: a vector of the initial mean likelihood densities, 4 * E_b/N_o.
                    Note that this SNR should be normalised using :func:`get_normalised_SNR` in :class:`PolarCode`
        :type z0: ndarray<float> or float
        :return: channel reliabilities in log-domain (least reliable first), and the frozen indices
        :rtype: ndarray<int>, ndarray<int>
        """

        z = np.zeros((myPC.N, myPC.n + 1))
        z[:, 0] = z0  # initial channel states

        for j in range(1, myPC.n + 1):
            u = 2 ** j  # number of branches at depth j
            # loop over top branches at this stage
            for t in range(0, myPC.N, u):
                for s in range(int(u / 2)):
                    k = t + s
                    z_top = z[k, j - 1]
                    z_bottom = z[k + int(u / 2), j - 1]

                    z[k, j] = self.phi_inv(1 - (1 - self.phi(z_top)) * (1 - self.phi(z_bottom)))
                    z[k + int(u / 2), j] = z_top + z_bottom

        m = np.array([self.logQ_Borjesson(0.707*np.sqrt(z[i, myPC.n])) for i in range(myPC.N)])
        reliabilities = np.argsort(-m, kind='mergesort')   # ordered by least reliable to most reliable
        frozen = np.argsort(m, kind='mergesort')[myPC.K:]     # select N-K least reliable channels
        return reliabilities, frozen



"""
Polar Codes in Python
=============================================
"""

from . import Math
from . import Construct
from . import Shorten
from . import Encode
from . import Decode
from . import AWGN
from . import Puncture

__all__ = [
    'Math',
    'Construct',
    'Shorten',
    'Encode',
    'Decode',
    'AWGN',
    'Puncture'
]



from tkinter import *
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
import numpy as np

class GUI:
    def __init__(self, master, myPC):
        # initialise GUI global vars
        self.master = master
        self.my_widget_ptrs = []
        self.fig = Figure(figsize=(10, 5), dpi=60)
        self.new_plot = self.fig.add_subplot(111)
        self.myPC = myPC

        # generate section frames
        first_frame = self.make_inputs()
        footer_frame = self.make_footer()
        self.myPC.gui_widgets = self.my_widget_ptrs

        # section layout
        first_frame.pack()
        self.make_plot(self.fig)
        footer_frame.pack(side=BOTTOM, fill=X)

    def make_plot(self, fig):
        # add plot to tkinter canvas
        self.canvas = FigureCanvasTkAgg(fig, master=self.master)
        self.canvas.draw()
        graph_widget = self.canvas.get_tk_widget()
        graph_widget.pack(pady=10)
        return graph_widget

    def make_inputs(self):
        my_frame = Frame(self.master)

        # labels
        titleFont = ("Arial", 10, "bold")
        pc_title_label = Label(my_frame, text="Polar Code", font=titleFont)
        puncturing_title_label = Label(my_frame, text="Puncturing/Shortening", font=titleFont)
        sim_label = Label(my_frame, text="Simulation", font=titleFont)
        plot_label = Label(my_frame, text="Plot", font=titleFont)
        n_label = Label(my_frame, text="N")
        k_label = Label(my_frame, text="K")
        const_title_label = Label(my_frame, text="Mothercode Construction")
        design_label = Label(my_frame, text="Design Eb/No (dB)")
        frozen_label = Label(my_frame, text="Frozen Bits")
        punct_algorithm_label = Label(my_frame, text="Algorithm")
        punctured_label = Label(my_frame, text="Punctured/Shortened Bits")
        iter = Label(my_frame, text="Iterations")
        min_frame_errors = Label(my_frame, text="Min. Frame Errors")
        save_dir_label = Label(my_frame, text="Directory")
        filenames_label = Label(my_frame, text="Filename(s)")
        save_label = Label(my_frame, text="Save to")
        Eb_No_values_label = Label(my_frame, text="Eb/No Values (dB)")

        # entries and text fields
        n_entry = Entry(my_frame)
        k_entry = Entry(my_frame)
        design_entry = Entry(my_frame)
        iter_entry = Entry(my_frame)
        min_frame_errors_entry = Entry(my_frame)
        file_dir_entry = Entry(my_frame)
        filenames_entry = Entry(my_frame)
        savefilename_entry = Entry(my_frame)
        frozen_bits = Text(my_frame, width=30, height=5)
        shortened_bits = Text(my_frame, width=30, height=5)
        Eb_No_values_entry = Entry(my_frame)

        n_entry.insert(END, '100')
        k_entry.insert(END, '64')
        design_entry.insert(END, '5')
        iter_entry.insert(END, '10000')
        min_frame_errors_entry.insert(END, '30')
        savefilename_entry.insert(END, 'data/pc_sim')
        file_dir_entry.insert(END, 'data/')
        filenames_entry.insert(END, 'pc_sim')
        Eb_No_values_entry.insert(END, '1,2,3,4')

        # menus
        const_options = ['bb', 'ga']
        const_opt_val = StringVar(my_frame)
        const_opt_val.set(const_options[0])
        const_opt = OptionMenu(my_frame, const_opt_val, *const_options)

        shortening_options = ['brs', 'wls', 'bgl']
        shortening_opt_val = StringVar(my_frame)
        shortening_opt_val.set(shortening_options[0])
        shortening_opt = OptionMenu(my_frame, shortening_opt_val, *shortening_options)

        # check and radio buttons
        manual_construction_flag = IntVar()
        punct_type_var = IntVar()
        punct_type_var.set(1)

        c = Checkbutton(my_frame, text="Manual Construction", variable=manual_construction_flag)
        p_opt = Radiobutton(my_frame, text="Puncturing", value=0, variable=punct_type_var)
        s_opt = Radiobutton(my_frame, text="Shortening", value=1, variable=punct_type_var)

        # buttons
        button_1 = Button(my_frame, text="Simulate",
                          command=lambda: [self.update_dict(), self.myPC.gui_sim_handler(self.dict)])
        button_2 = Button(my_frame, text="Plot",
                          command=lambda: [self.update_dict(), self.myPC.gui_plot_handler(self.dict, self.new_plot),
                                           self.canvas.draw()])

        # grid layout
        pc_title_label.grid(row=0, columnspan=2)
        n_label.grid(row=1, sticky=E)
        k_label.grid(row=2, sticky=E)
        n_entry.grid(row=1, column=1, sticky="nsew")
        k_entry.grid(row=2, column=1, sticky="nsew")
        design_label.grid(row=5, column=0, sticky=E)
        design_entry.grid(row=5, column=1, sticky="nsew")
        frozen_bits.grid(row=4, column=1)
        frozen_label.grid(row=4, column=0, sticky=E)
        const_title_label.grid(row=3, column=0, sticky=E)
        const_opt.grid(row=3, column=1, sticky="nsew")
        c.grid(row=6, column = 1)
        puncturing_title_label.grid(row=0, column=2, columnspan=2)
        shortening_opt.grid(row=2, column=3, sticky="nsew")
        punct_algorithm_label.grid(row=2, column=2, sticky=E)
        p_opt.grid(row=1, column=2)
        s_opt.grid(row=1, column=3)
        shortened_bits.grid(row=4, column=3)
        punctured_label.grid(row=4, column=2, sticky=E)
        save_label.grid(row=11, column=0, sticky=E)
        savefilename_entry.grid(row=11, column=1, sticky="nsew")
        save_dir_label.grid(row=9, column=2, sticky=E)
        file_dir_entry.grid(row=9, column=3, sticky="nsew")
        sim_label.grid(row=8, columnspan=2, column=0)
        plot_label.grid(row=8, columnspan=2, column=2)
        iter_entry.grid(row=9, column=1, sticky="nsew")
        min_frame_errors_entry.grid(row=10, column=1, sticky="nsew")
        iter.grid(row=9, column=0, sticky=E)
        min_frame_errors.grid(row=10, column=0, sticky=E)
        filenames_label.grid(row=10, column=2, sticky=E)
        filenames_entry.grid(row=10, column=3, sticky="nsew")
        Eb_No_values_label.grid(row=12, column=0, sticky=E)
        Eb_No_values_entry.grid(row=12, column=1, sticky="nsew")
        button_1.grid(row=13, column=0, columnspan=2, sticky="nsew")
        button_2.grid(row=13, column=2, columnspan=4, sticky="nsew")

        # save data widget pointers
        my_widget_ptrs = [n_entry, k_entry, const_opt_val, frozen_bits, punct_type_var, shortening_opt_val, shortened_bits, iter_entry,
                          min_frame_errors_entry, savefilename_entry, file_dir_entry, filenames_entry, design_entry, manual_construction_flag,
                          Eb_No_values_entry]
        self.my_widget_ptrs = my_widget_ptrs
        return my_frame

    def make_footer(self):
        my_frame = Frame(self.master)
        my_text_var = StringVar(my_frame)
        my_text_var.set("Simulation updates here...")
        status = Label(my_frame, relief=SUNKEN, anchor=W, textvariable=my_text_var)
        status.pack(fill=X)
        self.myPC.status_bar = my_text_var
        return my_frame

    def update_dict(self):
        # create GUI parameters object
        self.dict = {
            'N': int(self.my_widget_ptrs[0].get()),
            'K': int(self.my_widget_ptrs[1].get()),
            'construction_algo': self.my_widget_ptrs[2].get(),
            'frozen_set': [int(x) for x in self.my_widget_ptrs[3].get("1.0", END).strip().split(",") if x != ''],
            'punct_type': int(self.my_widget_ptrs[4].get()),
            'punct_algo': self.my_widget_ptrs[5].get(),
            'shortened_set': [int(x) for x in self.my_widget_ptrs[6].get("1.0", END).strip().split(",") if x != ''],
            'iterations': int(self.my_widget_ptrs[7].get()),
            'min_frame_errors': int(self.my_widget_ptrs[8].get()),
            'save_to': self.my_widget_ptrs[9].get(),
            'file_dir': self.my_widget_ptrs[10].get(),
            'filenames': self.my_widget_ptrs[11].get().strip().replace(' ', '').split(","),
            'design_SNR': int(self.my_widget_ptrs[12].get()),
            'manual_const_flag': self.my_widget_ptrs[13].get(),
            'snr_values': np.array([int(x) for x in self.my_widget_ptrs[14].get().strip().split(',')])
        }



import numpy as np
from polarcodes.Math import Math
from polarcodes.Construct import Construct

class Puncture(Construct):
    def __init__(self, myPC, design_SNR, manual=False):
        """
        In the future, this class will contain common puncturing algorithms.

        :param myPC:
        :param design_SNR:
        :param manual:
        """

        super().__init__(myPC, design_SNR, True)
        if manual:
            return


#!/usr/bin/env python

"""
This class simulates an AWGN channel by adding gaussian noise with double-sided noise power.
It updates ``likelihoods`` in :class:`PolarCode` with randomly generated log-likelihood ratios
for ``u`` in :class:`PolarCode`. For puncturing, the likelihoods for the punctured bits given by
``source_set_lookup`` in :class:`PolarCode` will be set to zero. For shortening,
these likelihoods will be set to infinity. Currently only BPSK modulation is supported.
"""

import matplotlib.pyplot as plt
import numpy as np

class AWGN:
    def __init__(self, myPC, Eb_No, plot_noise = False):
        """
        :param myPC: a polar code object created using the :class:`PolarCode` class
        :param Eb_No: the design SNR in decibels
        :param plot_noise: a flag to view the modeled noise
        :type myPC: :class:`PolarCode`
        :type Eb_No: float
        :type plot_noise: bool
        """

        self.myPC = myPC
        self.Es = myPC.get_normalised_SNR(Eb_No)
        self.No = 1
        self.plot_noise = plot_noise

        tx = self.modulation(self.myPC.u)
        rx = tx + self.noise(self.myPC.N)
        self.myPC.likelihoods = np.array(self.get_likelihoods(rx), dtype=np.float64)

        # change shortened/punctured bit LLRs
        if self.myPC.punct_flag:
            if self.myPC.punct_type == 'shorten':
                self.myPC.likelihoods[self.myPC.source_set_lookup == 0] = np.inf
            elif self.myPC.punct_type == 'punct':
                self.myPC.likelihoods[self.myPC.source_set_lookup == 0] = 0

    def LLR(self, y):
        """
        Finds the log-likelihood ratio of a received signal.
        LLR = Pr(y=0)/Pr(y=1).

        :param y: a received signal from a gaussian-distributed channel
        :type y: float
        :return: log-likelihood ratio for the input signal ``y``
        :rtype: float
        """

        return -2 * y * np.sqrt(self.Es) / self.No

    def get_likelihoods(self, y):
        """
        Finds the log-likelihood ratio of an ensemble of received signals using :func:`LLR`.

        :param y: an ensemble of received signals
        :type y: ndarray<float>
        :return: log-likelihood ratios for the input signals ``y``
        :rtype: ndarray<float>
        """
        return [self.LLR(y[i]) for i in range(len(y))]

    def modulation(self, x):
        """
        BPSK modulation for a bit field.
        "1" maps to +sqrt(E_s) and "0" maps to -sqrt(E_s).

        :param x: an ensemble of information to send
        :type x: ndarray<int>
        :return: Modulated signal with the information from ``x``
        :rtype: ndarray<float>
        """

        return 2 * (x - 0.5) * np.sqrt(self.Es)

    def noise(self, N):
        """
        Generate gaussian noise with a specified noise power.
        For a noise power N_o, the double-side noise power is N_o/2.

        :param N: the noise power
        :type N: float
        :return: white gaussian noise vector
        :rtype: ndarray<float>
        """

        # gaussian RNG vector
        s = np.random.normal(0, np.sqrt(self.No / 2), size=N)

        # display RNG values with ideal gaussian pdf
        if self.plot_noise:
            num_bins = 1000
            count, bins, ignored = plt.hist(s, num_bins, density=True)
            plt.plot(bins, 1 / (np.sqrt(np.pi * self.No)) * np.exp(- (bins) ** 2 / self.No),
                        linewidth=2, color='r')
            plt.title('AWGN')
            plt.xlabel('Noise, n')
            plt.ylabel('Density')
            plt.legend(['Theoretical', 'RNG'])
            plt.draw()
        return s

    def show_noise(self):
        """
        Trigger showing the gaussian noise. Only works if ``plot_noise`` is True.
        """
        plt.show()


#!/usr/bin/env python

"""
A polar encoder class. Currently only non-systematic encoding is supported.
"""

import numpy as np
from polarcodes.Math import Math

class Encode(Math):
    def __init__(self, myPC, encoder_name = 'polar_encode'):
        """
        :param myPC: a polar code object created using the :class:`PolarCode` class
        :param encoder_name: the name of the polar encoder implementation.
                            'polar_encode' => a non_recursive implementation (default).
                            'polar_encode_recursive' => a recursive implementation.
        :type myPC: :class:`PolarCode`
        :type encoder_name: string
        """

        self.myPC = myPC
        if encoder_name == 'polar_encode':
            self.polar_encode()
        elif encoder_name == 'polar_encode_recursive':
            self.polar_encode2(0, myPC.N-1)

    def polar_encode2(self, i1, i2):
        """
        Encodes a message using polar coding with a recursive implementation.
        The message ``x`` is encoded using in-place operations of output ``u`` in ``myPC``.
        The initial call of :func:`polar_encode2` should set (i1, i2) = (0, N-1) for a block length N.
        For example, the second partition indices will be (0, N/2-1) and (N/2, N-1).

        :param i1: start index of the partition
        :param i2: end index of the partition
        :type i1: int
        :type i2: int
        """
        h_shift = int((i2 - i1 + 1) / 2)  # length of each new partition
        mid = i1 + h_shift  # right-aligned mid-point

        for k in range(i1, mid):
            self.myPC.u[k] = self.myPC.u[k] ^ self.myPC.u[k + h_shift]  # add right partition to the left partition (mod 2)

        if h_shift >= 2:
            self.polar_encode2(i1, mid - 1)     # recursion on left partition
            self.polar_encode2(mid, i2)         # recursion on right partition

    # Polar encode x without using recursion
    # x = message bit field
    def polar_encode(self):
        """
        Encodes a message using polar coding with a non-recursive implementation.
        The message ``x`` is encoded using in-place operations of output ``u`` in ``myPC``.
        """

        # loop over the M stages
        n = self.myPC.N
        for i in range(self.myPC.N):
            if n == 1:  # base case: when partition length is 1
                break
            n_split = int(n / 2)

            # select the first index for each split partition, we always have n=2*n_split
            for p in range(0, self.myPC.N, n):
                # loop through left partitions and add the right partitions for a previous partition
                for k in range(n_split):
                    l = p + k
                    self.myPC.u[l] = self.myPC.u[l] ^ self.myPC.u[l + n_split]
            n = n_split


#!/usr/bin/env python

"""
An object that encapsulates all of the parameters required to define a polar code.
This object must be given to the following classes: :class:`AWGN`, :class:`Construct`, :class:`Decode`,
:class:`Encode`, :class:`GUI`, :class:`Shorten`.
"""

import numpy as np
from polarcodes.Math import Math
from polarcodes.Construct import Construct
from polarcodes.Shorten import Shorten
from polarcodes.Encode import Encode
from polarcodes.Decode import Decode
from polarcodes.AWGN import AWGN
import json
import matplotlib.pyplot as plt
import threading
import tkinter as tk

class PolarCode(Math):
    """
    :var N: the mothercode block length
    :var M: the block length (after puncturing)
    :var K: the code dimension
    :var n: number of bits per index
    :var s: number of shortened bit-channels
    :var reliabilities: reliability vector (least reliable to most reliable)
    :var frozen: the frozen bit indices
    :var frozen_lookup: lookup table for the frozen bits
    :var x: the uncoded message with frozen bits
    :var construction_type: the mothercode construction type
    :var message_received: the decoded message received from a channel
    :var punct_flag: whether or not the code is punctured
    :var simulated_snr: the SNR values simulated
    :var simulated_fer: the FER values for the SNR values in ``simulated_snr`` using :func:`simulate`
    :var simulated_ber: the BER values for the SNR values in ``simulated_snr`` using :func:`simulate`
    :var punct_type: 'punct' for puncturing, and 'shorten' for shortening
    :var punct_set: the coded punctured indices
    :var punct_set_lookup: lookup table for ``punct_set``
    :var source_set: the uncoded punctured indices
    :var source_set_lookup: lookup table for ``source_set``
    :var punct_algorithm: the name of a puncturing algorithm. Options: {'brs', 'wls', 'bgl', 'perm'}
    :var update_frozen_flag: whether or not to update the frozen indices after puncturing
    :var recip_flag: True if ``punct_set`` equals ``source_set``
    """

    def __init__(self, M, K, punct_params=('', '', [], [], [],)):
        """
        :param M: the block length (after puncturing)
        :param K: the code dimension
        :param punct_params: a tuple to completely specify the puncturing parameters (if required).
                            The syntax is (``punct_type``, ``punct_algorithm``, ``punct_set``, ``source_set``, ``update_frozen_flag``)
        :type M: int
        :type K: int
        :type punct_params: tuple
        """

        self.initialise_code(M, K, punct_params)
        self.status_bar = None  # set by the GUI so that the simulation progress can be tracked
        self.gui_widgets = []

    def initialise_code(self, M, K, punct_params):
        """
        Initialise the code with a set of parameters the same way as the constructor.
        Call this any time you want to change the code rate.
        """

        # mothercode parameters
        self.M = M
        self.N = int(2**(np.ceil(np.log2(M))))
        self.n = int(np.log2(self.N))
        self.K = K
        self.s = self.N - self.M
        self.reliabilities = np.array([])
        self.frozen = np.array([])
        self.frozen_lookup = np.array([])
        self.x = np.zeros(self.N, dtype=int)
        self.construction_type = 'bb'
        self.message_received = np.array([])
        self.punct_flag = False if self.M == self.N else True
        self.simulated_snr = np.array([])
        self.simulated_fer = np.array([])
        self.simulated_ber = np.array([])

        # puncturing parameters
        self.punct_type = punct_params[0]
        self.punct_set = np.array(punct_params[2])
        self.punct_set_lookup = self.get_lut(punct_params[2])
        self.source_set = np.array(punct_params[3])
        self.source_set_lookup = self.get_lut(punct_params[3])
        self.punct_algorithm = punct_params[1]
        self.update_frozen_flag = punct_params[4]
        self.recip_flag = np.array_equal(np.array(punct_params[2]), np.array(punct_params[3]))

    def __str__(self):
        """
        A string definition of PolarCode. This allows you to print any PolarCode object and see all of its
        relevant parameters.

        :return: a stringified version of PolarCode
        :rtype: string
        """

        output = '=' * 10 + " Polar Code " + '=' * 10 + '\n'
        output += "N: " + str(self.N) + '\n'
        output += "M: " + str(self.M) + '\n'
        output += "K: "+ str(self.K) + '\n'
        output += "Mothercode Construction: " + self.construction_type + '\n'
        output += "Ordered Bits (least reliable to most reliable): " + str(self.reliabilities) + '\n'
        output += "Frozen Bits: " + str(self.frozen) + '\n'
        output += "Puncturing Flag: " + str(self.punct_flag) + '\n'
        output += "Puncturing Parameters: {punct_type: " + str(self.punct_type) + '\n'
        output += "                        punct_algorithm: " + str(self.punct_algorithm) + '\n'
        output += "                        punct_set: " + str(self.punct_set) + '\n'
        output += "                        source_set: " + str(self.source_set) + '\n'
        output += "                        update_frozen_flag: " + str(self.update_frozen_flag) + "}" + '\n'
        return output

    def set_message(self, m):
        """
        Set the message vector to the non-frozen bits in ``x``. The frozen bits in ``frozen`` are set to zero.

        :param m: the message vector
        :type m: ndarray<int>
        """

        self.message = m
        self.x[self.frozen_lookup == 1] = m
        self.u = self.x.copy()

    def get_normalised_SNR(self, design_SNR):
        """
        Normalise E_b/N_o so that the message bits have the same energy for any code rate.

        :param design_SNR: E_b/N_o in decibels
        :type design_SNR: float
        :return: normalised E_b/N_o in linear units
        :rtype: float
        """

        Eb_No_dB = design_SNR
        Eb_No = 10 ** (Eb_No_dB / 10)  # convert dB scale to linear
        Eb_No = Eb_No * (self.K / self.M)  # normalised message signal energy by R=K/M (M=N if not punctured)
        return Eb_No

    def get_lut(self, my_set):
        """
        Convert a set into a lookup table.

        :param my_set: a vector of indices
        :type my_set: ndarray<int>
        :return: lookup table. "0" for an index in ``my_set``, else "1"
        :rtype: ndarray<int>
        """

        my_lut = np.ones(self.N, dtype=int)
        my_lut[my_set] = 0
        return my_lut

    def save_as_json(self, sim_filename):
        """
        Save all the important parameters in this object as a JSON file.

        :param sim_filename: directory and filename to save JSON file to (excluding extension)
        :type sim_filename: string
        """
        data = {
            'N': self.M,
            'n': self.n,
            'K': self.K,
            'frozen': self.frozen.tolist(),
            'construction_type': self.construction_type,
            'punct_flag': self.punct_flag,
            'punct_type': self.punct_type,
            'punct_set': self.punct_set.tolist(),
            'source_set': self.source_set.tolist(),
            'punct_algorithm': self.punct_algorithm,
            'update_frozen_flag': self.update_frozen_flag,
            'BER': self.simulated_ber.tolist(),
            'FER': self.simulated_fer.tolist(),
            'SNR': self.simulated_snr.tolist()
        }
        with open(sim_filename + '.json', 'w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

    def run_simulation(self, Eb_No, max_iter, min_errors, min_iters):
        frame_error_count = 0
        bit_error_count = 0
        num_blocks = 0
        for i in range(1, max_iter + 1):
            # simulate random PC in an AWGN channel
            self.set_message(np.random.randint(2, size=self.K))
            Encode(self)
            AWGN(self, Eb_No)
            Decode(self)

            # detect errors
            error_vec = self.message ^ self.message_received
            num_errors = sum(error_vec)
            frame_error_count = frame_error_count + (num_errors > 1)
            bit_error_count = bit_error_count + num_errors

            # early stopping condition
            num_blocks = i
            if frame_error_count >= min_errors and i >= min_iters:
                break
        return frame_error_count, bit_error_count, num_blocks

    def simulate(self, save_to, Eb_No_vec, design_SNR=None, max_iter=100000, min_iterations=1000, min_errors=30, sim_seed=1729, manual_const_flag=True):
        """
        Monte-carlo simulation of the performance of this polar code.
        The simulation has an early stopping condition of when the number of errors is below min_errors.
        Each E_b/N_o simulation has an additional early stopping condition using the minimum iterations
        and the minimum number of errors. The results are saved in a JSON file using :func:`save_as_json`.

        :param save_to: directory and filename to save JSON file to (excluding extension)
        :param Eb_No_vec: the range of SNR values to simulate
        :param design_SNR: the construction design SNR, E_b/N_o
        :param max_iter: maximum number of iterations per SNR
        :param min_iterations: the minimum number of iterations before early stopping is allowed per SNR
        :param min_errors: the minimum number of frame errors before early stopping is allowed per SNR
        :param sim_seed: pseudo-random generator seed, default is 1729 ('twister' on MATLAB)
        :param manual_const_flag: a flag that decides if construction should be done before simulating.
                                Set to False if mothercode and/or puncturing constructions are manually set by the user.
        :type save_to: string
        :type Eb_No_vec: ndarray<float>
        :type design_SNR: float
        :type max_iter: int
        :type min_iterations: int
        :type min_errors: int
        :type sim_seed: int
        :type manual_const_flag: bool
        """

        # initialise simulation
        np.random.seed(sim_seed)
        frame_error_rates = np.zeros(len(Eb_No_vec))
        bit_error_rates = np.zeros(len(Eb_No_vec))

        # do construction if not done already
        if not manual_const_flag:
            if self.punct_flag and self.punct_type == 'shorten':
                Shorten(self, design_SNR)
                print('SHORTEN')
            else:
                print('MC')
                Construct(self, design_SNR)

        print(self)
        print('=' * 10, "Simulation", '=' * 10)
        for i in range(len(Eb_No_vec)):
            # run simulation for the current SNR
            frame_error_count, bit_error_count, num_blocks = self.run_simulation(Eb_No_vec[i], max_iter, min_errors, min_iterations)

            # calculate FER and BER
            frame_error_rate = frame_error_count / num_blocks
            bit_error_rate = bit_error_count / (self.K * num_blocks)
            frame_error_rates[i] = frame_error_rate
            bit_error_rates[i] = bit_error_rate
            print("Eb/No:", round(Eb_No_vec[i], 5), "  FER:", round(frame_error_rate, 3), "  BER:", round(bit_error_rate, 5))
            print('# Iterations:', num_blocks, '  # Frame Errors:', frame_error_count, ' # Bit Errors:', bit_error_count)
            print('='*20)

            # update GUI (if used)
            if self.status_bar != None:
                self.status_bar.set("Simulation progress: " + str(i + 1) + "/" + str(len(Eb_No_vec)))

            # early stopping condition
            if frame_error_count < min_errors:
                break

        # write data to JSON file
        self.simulated_snr = Eb_No_vec
        self.simulated_ber = bit_error_rates
        self.simulated_fer = frame_error_rates
        self.save_as_json(save_to)

        # update GUI construction fields (if used)
        if self.status_bar != None:
            self.gui_widgets[3].delete("1.0", tk.END)
            self.gui_widgets[6].delete("1.0", tk.END)
            self.gui_widgets[3].insert(tk.INSERT, ",".join(map(str, self.frozen)))
            self.gui_widgets[6].insert(tk.INSERT, ",".join(map(str, self.punct_set)))

        # update console and GUI
        print("Successfully completed simulation.\n")
        if self.status_bar != None:
            self.status_bar.set("Simulation progress: Done.")

    def plot_helper(self, new_plot, sim_filenames, dir, plot_title = 'Polar Code Performance'):
        # plot the FER and BER from file list
        new_plot.cla()
        for sim_filename in sim_filenames:
            with open(dir + sim_filename + '.json') as data_file:
                data_loaded = json.load(data_file)
            new_plot.plot(data_loaded['SNR'], data_loaded['FER'], '-o', markersize=6, linewidth=3, label=sim_filename)

        # format the plots
        new_plot.set_title(plot_title)
        new_plot.set_ylabel("Frame Error Rate")
        new_plot.set_xlabel("$E_b/N_o$ (dB)")
        new_plot.grid(linestyle='-')
        new_plot.set_yscale('log')
        new_plot.legend(loc='lower left')

    # call this for manual plotting
    def plot(self, sim_filenames, dir):
        """
        Plot multiple sets of FER data from the same directory on the same axes.

        :param sim_filenames: a list of all filenames to plot in a common root directory
        :param dir: the root directory for the specified filenames
        :type sim_filenames: ndarray<string>
        :type dir: string
        """

        fig = plt.figure()
        new_plot = fig.add_subplot(111)
        self.plot_helper(new_plot, sim_filenames, dir)
        fig.show()

    # used by the GUI class for automated plotting
    def gui_plot_handler(self, gui_dict, fig):
        sim_filenames = gui_dict['filenames']
        dir = gui_dict['file_dir']
        self.plot_helper(fig, sim_filenames, dir)

    # used by the GUI class for simulating a new code
    def gui_sim_handler(self, gui_dict):
        # updated Polar Code from user
        punct_type = 'shorten' if gui_dict['punct_type'] == True else 'punct'
        shortening_params = (punct_type, gui_dict['punct_algo'], np.array(gui_dict['shortened_set'], dtype=int),
                             np.array(gui_dict['shortened_set'], dtype=int), False)
        self.initialise_code(gui_dict['N'], gui_dict['K'], shortening_params)
        self.construction_type = gui_dict['construction_algo']
        self.frozen = gui_dict['frozen_set']

        # simulation parameters from user
        iterations = gui_dict['iterations']
        min_frame_errors = gui_dict['min_frame_errors']
        file_dir = gui_dict['file_dir']
        save_to = gui_dict['save_to']
        manual_const_flag = gui_dict['manual_const_flag']
        design_SNR = gui_dict['design_SNR']
        Eb_No_vec = gui_dict['snr_values']

        # run simulation in another thread to avoid GUI freeze
        x = threading.Thread(name='sim_thread', target=self.simulate, args=(save_to, Eb_No_vec, design_SNR, iterations, 1000, min_frame_errors, 1729, manual_const_flag,))
        x.setDaemon(True)
        x.start()




#!/usr/bin/env python

"""
A class dedicated to shortening. This means that the likelihoods for each coded shortened bit are set to infinity at the channel output given by class AWGN.
Shortening techniques supported: Wang-Liu Shortening (WLS), Bit-Reversal Shortening (BRS), Bioglio-Gabry-Land Shortening (BGL), and Permuted WLS (PWLS).
"""

import numpy as np
from polarcodes.Math import Math
from polarcodes.Construct import Construct

class Shorten(Construct):
    def __init__(self, myPC, design_SNR, manual=False):
        """
        :param myPC: a polar code object created using the :class:`PolarCode` class
        :param design_SNR: the design SNR in decibels
        :param manual: suppress the constructor init
        :type myPC: :class:`PolarCode`
        :type design_SNR: float
        :type manual: bool
        """

        super().__init__(myPC, design_SNR, True)
        if manual:
            return
        else:
            self.update_spcc(myPC, design_SNR)

    def update_spcc(self, myPC, design_SNR):
        # select shortening construction method
        if myPC.punct_algorithm == 'brs':  # BRS shortening
            myPC.punct_set = self.brs_pattern(myPC)
            myPC.source_set = myPC.punct_set
        elif myPC.punct_algorithm == 'wls':  # WLS shortening
            myPC.punct_set = self.wls_pattern(myPC)
            myPC.source_set = myPC.punct_set
        elif myPC.punct_algorithm == 'bgl':  # BGL shortening
            myPC.punct_set = self.bgl_pattern(myPC)
            myPC.source_set = myPC.punct_set
        elif myPC.punct_algorithm == 'perm':  # Perm shortening
            myPC.punct_set = self.perm(myPC)
            myPC.source_set = myPC.punct_set

        myPC.punct_set_lookup = myPC.get_lut(myPC.punct_set)
        myPC.source_set_lookup = myPC.get_lut(myPC.source_set)

        # decide if we want a puncturing-dependent frozen set
        if not myPC.update_frozen_flag:
            self.update_mpcc(myPC, design_SNR)
        else:
            self.shortened_pcc(myPC, design_SNR)
        myPC.frozen = self.frozen_from_pattern(myPC)
        myPC.frozen_lookup = myPC.get_lut(myPC.frozen)

    def shortened_pcc(self, myPC, design_SNR):
        """
        Find the shortened polar code construction and update ``frozen`` in ``myPC``.
        This is not strictly necessary, since many shortening patterns work just fine with the mothercode reliabilities.

        :param myPC: a polar code object created using the :class:`PolarCode` class
        :param design_SNR: the design SNR in decibels
        """

        # select the construction method
        design_SNR_normalised = myPC.get_normalised_SNR(design_SNR)
        if myPC.construction_type == 'bb':
            z0 = np.array([-design_SNR_normalised] * myPC.N)
            z0[myPC.punct_set] = -np.inf
            myPC.reliabilities,_ = self.general_pcc(myPC, z0)
        elif myPC.construction_type == 'ga':
            z0 = np.array([4 * design_SNR_normalised] * myPC.N)
            z0[myPC.punct_set] = np.inf
            myPC.reliabilities,_ = self.general_ga(myPC, z0)

    def wls_pattern(self, myPC):
        """
        Wang-Liu Shortening (WLS). The common pattern from the Wang-Liu algorithm.

        -------------
        **References:**

        * Runxin Wang, & Rongke Liu. (2014). A Novel Puncturing Scheme for Polar Codes. IEEE Communications Letters, 18(12), 2081–2084. https://doi.org/10.1109/LCOMM.2014.2364845

        :param myPC: a polar code object created using the :class:`PolarCode` class
        :type myPC: :class:`PolarCode` class
        :return: WLS shortening set
        :rtype: ndarray<int>
        """

        punct_set = np.array(range(myPC.N - myPC.s, myPC.N))
        return punct_set

    def brs_pattern(self, myPC):
        """
        Bit-Reversal Shortening (BRS). A known high-performing shortening set, often called RQUP.

        -------------
        **References:**

        * Niu, Dai, Chen, Lin, Zhang, & Vasilakos. (2017). Rate-Compatible Punctured Polar Codes: Optimal Construction Based on Polar Spectra. arXiv.org. Retrieved from http://search.proquest.com/docview/2076458581/

        :param myPC: a polar code object created using the :class:`PolarCode` class
        :type myPC: :class:`PolarCode` class
        :return: BRS shortening set
        :rtype: ndarray<int>
        """
        punct_set_last = self.wls_pattern(myPC)
        punct_set = np.array([self.bit_reversed(i, myPC.n) for i in punct_set_last])
        return punct_set

    def frozen_from_pattern(self, myPC):
        """
        Forces the frozen bits to include the corresponding puncturing source bits in ``source_set`` in ``myPC``.

        :param myPC: a polar code object created using the :class:`PolarCode` class
        :type myPC: :class:`PolarCode` class
        :return: the new frozen set, that is typically assigned to ``frozen`` in ``myPC``.
        :rtype: ndarray<int>
        """

        R_m = []
        for i in range(myPC.N):    # add elements from R not in punct_set to R_m
            if myPC.reliabilities[i] not in myPC.source_set:
                R_m.append(myPC.reliabilities[i])
        t = myPC.M - myPC.K   # number of frozen bits left to select
        frozen = np.array(np.append(np.array(R_m[:t]), myPC.source_set))   # first t bits of R_m, then append S
        return frozen

    def perm(self, myPC):
        """
        Bit-wise permutation of the indices of the WLS pattern. This has been shown to produce other reciprocal
        shortening patterns, and so it is useful in enumerating them all for analysis.

        :param myPC: a polar code object created using the :class:`PolarCode` class
        :type myPC: :class:`PolarCode` class
        :return: the permuted shortening pattern for ``myPC``
        :rtype: ndarray<int>
        """

        punct_set_last = self.wls_pattern(myPC)
        punct_set = np.array(self.bit_perm(punct_set_last, myPC.perm, myPC.n))  # specify perm before construction
        return punct_set

    def wang_liu(self, myPC):
        N = 2**myPC.n
        G = self.arikan_gen(myPC.n)
        s = []

        for r in range(myPC.punct_set):
            for i in range(N):
                row = G[i, :]
                row_wt = np.sum(row)
                if row_wt == 1:
                    j = (np.where(row==1)[0]).item()
                    G[i, :] = np.zeros(N)
                    G[:, j] = np.zeros(N)
                    s.append(i)
                    break
        return np.array(s)

    def bgl_pattern(self, myPC):
        """
        Bioglio-Gabry-Land (BGL) Shortening

        -------------
        **References:**

        * Bioglio, V., Gabry, F., & Land, I. (2017). Low-Complexity Puncturing and Shortening of Polar Codes. arXiv.org. Retrieved from http://search.proquest.com/docview/2075581442/

        :param myPC: a polar code object created using the :class:`PolarCode` class
        :type myPC: :class:`PolarCode` class
        :return: the BGL shortening set
        :rtype: ndarray<int>
        """

        num_bits = myPC.n
        s = myPC.N - myPC.M  # number bits to shorten
        reversed_indices = np.array([self.bit_reversed(i, num_bits) for i in myPC.reliabilities])
        punct_set = np.array(reversed_indices[-s:])  # last s bits of reversed_indices
        return punct_set